<?php
// This file is part of the Studyplan plugin for Moodle
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.
/**
 * Edit/manage study plans
 * @package    local_treestudyplan
 * @copyright  2023 P.M. Kuipers
 * @license    https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

require_once("../../config.php");

require_once($CFG->libdir.'/weblib.php');

use local_treestudyplan\contextinfo;
use local_treestudyplan\studyplan;
use local_treestudyplan\courseservice;
use local_treestudyplan\debug;

$systemcontext = context_system::instance();
$PAGE->set_context($systemcontext);
$PAGE->set_url("/local/treestudyplan/edit-plan.php", []);
require_login();

// Figure out the context (category or system, based on either category or context parameter).
$categoryid = optional_param('categoryid', 0, PARAM_INT); // Category id.
$contextid = optional_param('contextid', 0, PARAM_INT); // Context id.
if ($categoryid > 0) {
    $studyplancontext = context_coursecat::instance($categoryid);

} else if ($contextid > 0) {
    $studyplancontext = context::instance_by_id($contextid);
    if (in_array($studyplancontext->contextlevel, [CONTEXT_SYSTEM, CONTEXT_COURSECAT])) {
        $categoryid = $studyplancontext->instanceid;
    } else {
        $studyplancontext = $systemcontext;
    }
} else {
    // If no context is selected, find the first available context for this user.
    $availablecontexts = courseservice::list_available_categories("edit");
    $contextid = 1; // Fallback to system context.
    if (count($availablecontexts) > 0) {
        $contextid = $availablecontexts[0]["context_id"];
    }
    // Reload page with selected category.
    $url = new \moodle_url('/local/treestudyplan/edit-plan.php', ["contextid" => $contextid]);
    header('Location: '.$url->out(false), true, 302);
    exit;
}

$ci = new contextinfo($studyplancontext);
$contextname = $ci->pathstr();

$PAGE->set_pagelayout('base');
$PAGE->set_title(get_string('cfg_plans', 'local_treestudyplan')." - ".$contextname);
$PAGE->set_heading($contextname);

if ($studyplancontext->id > 1) {
    navigation_node::override_active_url(new moodle_url('/course/index.php', ['categoryid' => $categoryid ]));
    $PAGE->navbar->add(get_string('cfg_plans', 'local_treestudyplan'));

    // Coursecat context.
    $cat = \core_course_category::get($studyplancontext->instanceid, IGNORE_MISSING, true); // We checck visibility later.
} else {
    // System context.
    $cat = \core_course_category::top();
}

if (!$cat->is_uservisible()) {

    throw new \moodle_exception(
        "error:cannotviewcategory",
        "local_treestudyplan",
        "/local/treestudyplan/edit_plan.php",
        $contextname
    );
}

if (!has_capability('local/treestudyplan:editstudyplan', $studyplancontext)) {
    throw new \moodle_exception(
        "error:nostudyplaneditaccess",
        "local_treestudyplan",
        "/local/treestudyplan/edit_plan.php",
        $contextname
    );
}

// Load javascripts and specific css.
$PAGE->requires->css(new moodle_url($CFG->wwwroot.'/local/treestudyplan/css/bootstrap-vue/bootstrap-vue.css'));
if ($CFG->debugdeveloper) {
    $PAGE->requires->css(new moodle_url($CFG->wwwroot.'/local/treestudyplan/css/devstyles.css'));
}
$PAGE->requires->js_call_amd('local_treestudyplan/page-edit-plan', 'init', [$studyplancontext->id, $categoryid, [
    "defaultAggregation" => get_config("local_treestudyplan", "aggregation_mode"),
]]);

print $OUTPUT->header();
$text = (object)[
    'loading' => get_string("loading", "core"),
    'studyplan_select' => get_string("studyplan_select", 'local_treestudyplan'),
    'studyplan_noneselected' => get_string("studyplan_noneselected", 'local_treestudyplan'),
    'defaultaggregation' => get_config("local_treestudyplan", "aggregation_mode"),
    'back' => get_string('back', 'local_treestudyplan'),
];

print <<<END
<div id='root' class="t-studyplan-limit-width">
    <div class='vue-loader' v-show='false'>
        <div class="spinner-border text-primary" role="status">
            <span class="sr-only">Loading...</span>
        </div>
    </div>
    <div v-cloak>
        <div v-if='!activestudyplan && usedcontexts && !loadingstudyplan' class='ml-3 mb-3 s-context-selector'>
            <b-form-select text='{$contextname}' :value="contextid" @change='switchContext'
                :class="(!(usedcontexts.length)) ? 'text-primary' : ''">
                <b-form-select-option v-if='!(usedcontexts.length)' :value="contextid"
                    :class="'text-primary'">
                    <span>{$text->loading}...</span></b-form-select-option>
                <b-form-select-option v-for='ctx in usedcontexts' :key='ctx.id' :value="ctx.context_id"
                    :class="(ctx.studyplancount > 0) ? 'font-weight-bold' : ''"
                ><span v-for="(p, i) in ctx.category.path"
                    ><span v-if="i>0"> / </span>{{ p }}</span> <span>({{ ctx.studyplancount }})</span>
                </b-form-select-option>
            </b-form-select>
            <div v-if="!(usedcontexts.length)"
                style="position: relative; top: 0.3rem; width: 1.2rem; height: 1.2rem; font-size: 0.7rem;"
                class="spinner-border text-primary" role="status"></div>

        </div>
        <h3 v-else>{$contextname}</h3>
        <div class="m-buttonbar" style="margin-bottom: 1em;">
            <a href='#' v-if='activestudyplan' @click.prevent='closeStudyplan'
                ><i style='font-size: 150%;' class='fa fa-chevron-left'></i> {$text->back}</a>
            <span v-if='activestudyplan'>{$text->studyplan_select}</span>&nbsp;
            <b-form-select v-if='activestudyplan' lazy :text='dropdown_title' :value='activestudyplan.id' @change="selectStudyplan">
                <b-form-select-option
                    v-for='(studyplan, planindex) in studyplans'
                    :value="studyplan.id"
                    :key='studyplan.id'
                    >{{ studyplan.name }}</b-form-select-option>
            </b-form-select>&nbsp;
            <template v-if='!activestudyplan && !loadingstudyplan'>
                <t-studyplan-edit
                    @created="onStudyPlanCreated"
                    mode="create"
                    v-model="create.studyplan"
                    type="button"
                    variant="primary"
                    default-aggregation="{$text->defaultaggregation}"
                    :contextid='contextid'
                ><i class='fa fa-plus'></i> {{ text.studyplan_add }}</t-studyplan-edit>
                <b-button
                        variant='danger' href='#' role='presentation' @click="import_studyplan "
                    ><i class='fa fa-upload'></i> {{ text.advanced_import_from_file }}</b-button
                ><mform v-if="premiumenabled && templatecount > 0"
                    name="studyplan_fromtemplateform"
                    :params="{contextid: contextid }"
                    @saved="onStudyPlanCreated"
                    variant="success"
                    type="button"
                    :title="text.advanced_import_from_file"
                    ><slot><i class='fa fa-clone'></i> {{ text.advanced_create_from_template }}</slot></mform>
            </template>
        </div>

        <div class='t-studyplan-container'>
            <t-studyplan
                v-if='activestudyplan'
                v-model='activestudyplan'
                @moved="movedStudyplan"
            ></t-studyplan>
            <div v-else-if='loadingstudyplan' class="spinner-border text-primary" role="status">
                <span class="sr-only">Loading...</span>
            </div>
            <div v-else class='t-studyplan-notselected'>
                <p>{$text->studyplan_noneselected}</p>
                <b-card-group deck>
                    <s-studyplan-card
                        v-for='(studyplan, planindex) in studyplans'
                        :key='studyplan.id'
                        v-model='studyplans[planindex]'
                        open
                        ignoresuspend
                        @open='selectStudyplan(studyplan.id)'
                        >
                        <template #title>
                            <span class='s-studyplan-card-title-buttons'>
                                <t-studyplan-edit
                                    v-model="studyplans[planindex]"
                                    @moved="movedStudyplan"
                                    @input="refreshTemplateCount"
                                ></t-studyplan-edit>
                                <t-studyplan-associate v-model="studyplans[planindex]"></t-studyplan-associate>
                            </span>
                        </template>
                    </s-studyplan-card>
                </b-card-group>
            </div>
        </div>
    </div>
</div>
END;
print $OUTPUT->footer();
